@@ -1,5 +1,6 @@
 """rxnmapper initialization."""
-__name__ = 'rxnmapper'
-__version__ = '0.1.0'
+__name__ = "rxnmapper"
+__version__ = "0.2.0"  # managed by bump2version
+
 
 from .core import RXNMapper  # noqa
