@@ -1,17 +1,23 @@
 """Contains functions needed to process reaction SMILES and their tokens"""
 from __future__ import absolute_import, division, print_function, unicode_literals
+
 import logging
-
 import re
+from functools import partial
+from typing import Any, List
+
 import numpy as np
-from functools import partial
-from typing import *
-from rdkit import Chem
-from rdkit.Chem import AllChem
-
-from functools import partial
-
-from rdkit import rdBase
+from rdkit import Chem, rdBase
+from rxn.chemutils.conversion import canonicalize_smiles, smiles_to_mol
+from rxn.chemutils.exceptions import InvalidSmiles
+from rxn.chemutils.reaction_equation import (
+    ReactionEquation,
+    apply_to_compounds,
+    merge_reactants_and_agents,
+    sort_compounds,
+)
+from rxn.chemutils.reaction_smiles import parse_any_reaction_smiles
+from rxn.chemutils.utils import remove_atom_mapping
 
 LOGGER = logging.getLogger("attnmapper:smiles_utils")
 
@@ -23,7 +29,7 @@ BAD_TOKS = ["[CLS]", "[SEP]"]  # Default Bad Tokens
 
 
 def tokenize(smiles: str) -> List[str]:
-    """ Tokenize a SMILES molecule or reaction """
+    """Tokenize a SMILES molecule or reaction"""
     regex = re.compile(SMI_REGEX_PATTERN)
     tokens = [token for token in regex.findall(smiles)]
     assert smiles == "".join(tokens)
@@ -31,23 +37,16 @@ def tokenize(smiles: str) -> List[str]:
 
 
 def get_atom_types(smiles: str):
-    """ Return atomic numbers for every token in (reaction) SMILES """
+    """Return atomic numbers for every token in (reaction) SMILES"""
     atom_tokens = get_atom_tokens_mask(smiles)
     if ">>" in smiles:
         precursors, products = smiles.split(">>")
 
-        precursors_mol = Chem.MolFromSmiles(precursors)
-        products_mol = Chem.MolFromSmiles(products)
-
-        atom_types = [
-            atom.GetAtomicNum() for atom in precursors_mol.GetAtoms()
-        ]
-        atom_types += [atom.GetAtomicNum() for atom in products_mol.GetAtoms()]
+        atom_types = get_atom_types_smiles(precursors)
+        atom_types += get_atom_types_smiles(products)
 
     else:
-        smiles_mol = Chem.MolFromSmiles(smiles)
-
-        atom_types = [atom.GetAtomicNum() for atom in smiles_mol.GetAtoms()]
+        atom_types = get_atom_types_smiles(smiles)
 
     token_types = []
     atom_counter = 0
@@ -63,14 +62,15 @@ def get_atom_types(smiles: str):
 
 def get_atom_types_smiles(smiles: str) -> List[int]:
     """Convert each atom in a SMILES into a list of their atomic numbers
-    
+
     Args:
-        smiles: SMILES representation of molecule
+        smiles: SMILES representation of a molecule or set of molecules.
 
     Returns:
         List of atom numbers for each atom in the smiles. Reports atoms in the same order they were passed in the original SMILES
     """
-    smiles_mol = Chem.MolFromSmiles(smiles)
+    # If `smiles` is a set of molecules, it may contain a "~".
+    smiles_mol = smiles_to_mol(smiles.replace("~", "."), sanitize=False)
 
     atom_types = [atom.GetAtomicNum() for atom in smiles_mol.GetAtoms()]
 
@@ -78,7 +78,7 @@ def get_atom_types_smiles(smiles: str) -> List[int]:
 
 
 def is_atom(token: str, special_tokens: List[str] = BAD_TOKS) -> bool:
-    """Determine whether a token is an atom. 
+    """Determine whether a token is an atom.
 
     Args:
         token: Token fed into the transformer model
@@ -93,8 +93,7 @@ def is_atom(token: str, special_tokens: List[str] = BAD_TOKS) -> bool:
     return (not is_bad) and normal_atom
 
 
-def number_tokens(tokens: List[str],
-                  special_tokens: List[str] = BAD_TOKS) -> List[int]:
+def number_tokens(tokens: List[str], special_tokens: List[str] = BAD_TOKS) -> List[int]:
     """Map list of tokens to a list of numbered atoms
 
     Args:
@@ -128,10 +127,10 @@ def get_graph_distance_matrix(smiles: str):
     Args:
         smiles {[type]} -- [description]
 
-    Returns: 
+    Returns:
         Numpy array representing the graphwise distance between each atom and every other atom in the molecular SMILES
     """
-    mol = Chem.MolFromSmiles(smiles)
+    mol = smiles_to_mol(smiles, sanitize=False)
     return Chem.GetDistanceMatrix(mol)
 
 
@@ -147,13 +146,13 @@ def get_adjacency_matrix(smiles: str):
         Equivalent to `distance_matrix[distance_matrix == 1]`
     """
 
-    mol = Chem.MolFromSmiles(smiles)
+    mol = smiles_to_mol(smiles, sanitize=False)
     return Chem.GetAdjacencyMatrix(mol)
 
 
 def is_mol_end(a: str, b: str) -> bool:
     """Determine if `a` and `b` are both tokens within a molecule (Used by the `group_with` function).
-    
+
     Returns False whenever either `a` or `b` is a molecule delimeter (`.` or `>>`)"""
     no_dot = (a != ".") and (b != ".")
     no_arrow = (a != ">>") and (b != ">>")
@@ -163,14 +162,14 @@ def is_mol_end(a: str, b: str) -> bool:
 
 def group_with(predicate, xs: List[Any]):
     """Takes a list and returns a list of lists where each sublist's elements are
-all satisfied pairwise comparison according to the provided function.
-Only adjacent elements are passed to the comparison function
+    all satisfied pairwise comparison according to the provided function.
+    Only adjacent elements are passed to the comparison function
 
-    Original implementation here: https://github.com/slavaGanzin/ramda.py/blob/master/ramda/group_with.py 
+        Original implementation here: https://github.com/slavaGanzin/ramda.py/blob/master/ramda/group_with.py
 
-    Args:
-        predicate ( f(a,b) => bool): A function that takes two subsequent inputs and returns True or Fale
-        xs: List to group
+        Args:
+            predicate ( f(a,b) => bool): A function that takes two subsequent inputs and returns True or Fale
+            xs: List to group
     """
     out = []
     is_str = isinstance(xs, str)
@@ -215,16 +214,13 @@ def tokens_to_adjacency(tokens: List[str]) -> np.array:
 
     Args:
         tokens: Tokenized SMILES representation
-        
+
     Returns:
         Numpy Array, where non-zero entries in row `i` indicate the tokens that are atom-adjacent to token `i`
     """
     from scipy.linalg import block_diag
 
     mol_tokens = split_into_mols(tokens)
-    altered_mol_tokens = [
-        m for m in mol_tokens if "." not in set(m) and ">>" not in set(m)
-    ]
 
     smiles = [
         tokens_to_smiles(mol, BAD_TOKS) for mol in mol_tokens
@@ -232,7 +228,7 @@ def tokens_to_adjacency(tokens: List[str]) -> np.array:
 
     # Calculate adjacency matrix for reaction
     altered_smiles = [
-        s for s in smiles if s not in set([".", ">>"])
+        s for s in smiles if s not in {".", "~", ">>"}
     ]  # Only care about atoms
     adjacency_mats = [
         get_adjacency_matrix(s) for s in altered_smiles
@@ -241,13 +237,12 @@ def tokens_to_adjacency(tokens: List[str]) -> np.array:
     return rxn_mask
 
 
-def get_mask_for_tokens(tokens: List[str],
-                        special_tokens: List[str] = []) -> List[int]:
+def get_mask_for_tokens(tokens: List[str], special_tokens: List[str] = []) -> List[int]:
     """Return a mask for a tokenized smiles, where atom tokens
     are converted to 1 and other tokens to 0.
-    
+
     e.g. c1ccncc1 would give [1, 0, 1, 1, 1, 1, 1, 0]
-    
+
     Args:
         smiles: Smiles string of reaction
         special_tokens: Any special tokens to explicitly not call an atom. E.g. "[CLS]" or "[SEP]"
@@ -261,14 +256,12 @@ def get_mask_for_tokens(tokens: List[str],
     return atom_token_mask
 
 
-def tok_mask(
-    tokens: List[str], special_tokens: List[str] = BAD_TOKS
-) -> np.array:
+def tok_mask(tokens: List[str], special_tokens: List[str] = BAD_TOKS) -> np.array:
     """Return a mask for a tokenized smiles, where atom tokens
     are converted to 1 and other tokens to 0.
-    
+
     e.g. c1ccncc1 would give [1, 0, 1, 1, 1, 1, 1, 0]
-    
+
     Args:
         smiles: Smiles string of reaction
         special_tokens: Any special tokens to explicitly not call an atom. E.g. "[CLS]" or "[SEP]"
@@ -283,9 +276,9 @@ def tok_mask(
 def get_atom_tokens_mask(smiles: str, special_tokens: List[str] = []):
     """Return a mask for a smiles, where atom tokens
     are converted to 1 and other tokens to 0.
-    
+
     e.g. c1ccncc1 would give [1, 0, 1, 1, 1, 1, 1, 0]
-    
+
     Args:
         smiles: Smiles string of reaction
         special_tokens: Any special tokens to explicitly not call an atom. E.g. "[CLS]" or "[SEP]"
@@ -297,10 +290,8 @@ def get_atom_tokens_mask(smiles: str, special_tokens: List[str] = []):
     return get_mask_for_tokens(tokens, special_tokens)
 
 
-def canonicalize_and_atom_map(
-    smi: str, return_equivalent_atoms=False
-):
-    """ Remove atom mapping, canonicalize and return mapping numbers in order of canonicalization. 
+def canonicalize_and_atom_map(smi: str, return_equivalent_atoms=False):
+    """Remove atom mapping, canonicalize and return mapping numbers in order of canonicalization.
 
     Args:
         smi: reaction SMILES str
@@ -309,9 +300,7 @@ def canonicalize_and_atom_map(
     Returns:
 
     """
-    mol = Chem.MolFromSmiles(smi)
-    if not mol:
-        raise NotCanonicalizableSmilesException("Molecule not canonicalizable")
+    mol = smiles_to_mol(smi, sanitize=False)
     for atom in mol.GetAtoms():
         if atom.HasProp("molAtomMapNumber"):
             atom_map = atom.GetAtomMapNum()
@@ -321,14 +310,12 @@ def canonicalize_and_atom_map(
             atom.SetProp("atom_map", str(0))
     can_smi = Chem.MolToSmiles(mol)
     order = list(
-        mol.GetPropsAsDict(includePrivate=True,
-                           includeComputed=True)["_smilesAtomOutputOrder"]
+        mol.GetPropsAsDict(includePrivate=True, includeComputed=True)[
+            "_smilesAtomOutputOrder"
+        ]
     )
 
-    atom_maps_canonical = [
-        mol.GetAtoms()[idx].GetProp("atom_map") for idx in order
-    ]
-
+    atom_maps_canonical = [mol.GetAtoms()[idx].GetProp("atom_map") for idx in order]
 
     if not return_equivalent_atoms:
         return (can_smi, atom_maps_canonical)
@@ -337,63 +324,73 @@ def canonicalize_and_atom_map(
 
 
 def generate_atom_mapped_reaction_atoms(
-    rxn: str, product_atom_maps, expected_atom_maps=None
-):
+    rxn: str, product_atom_maps, expected_atom_maps=None, canonical: bool = False
+) -> ReactionEquation:
     """
     Generate atom-mapped reaction from unmapped reaction and
-    product-2-reactant atoms mapping vector. 
+    product-2-reactant atoms mapping vector.
     Args:
-        rxn: unmapped reaction 
-        product_atom_maps: product to reactant atom maps
-        expected_atom_maps: (optional) if given return the differences
+        rxn: unmapped reaction, in the format that the transformer model relies on.
+        product_atom_maps: product to reactant atom maps.
+        expected_atom_maps: if given, return the differences.
+        canonical: whether to canonicalize the resulting SMILES.
 
     Returns: Atom-mapped reaction
 
     """
 
-    precursors, products = rxn.split(">>")
-    precursors_mol = Chem.MolFromSmiles(precursors)
-    products_mol = Chem.MolFromSmiles(products)
+    reactants, agents, products = parse_any_reaction_smiles(rxn)
+    precursors_mols = [smiles_to_mol(pr, sanitize=False) for pr in reactants + agents]
+    products_mols = [smiles_to_mol(prod, sanitize=False) for prod in products]
 
     precursors_atom_maps = []
 
     differing_maps = []
 
-    product_mapping_dict = {
+    product_mapping_dict = {}
 
-    }
+    i = -1
+    atom_mapped_precursors_list = []
+    for precursor_mol in precursors_mols:
+        for atom in precursor_mol.GetAtoms():
+            i += 1
+            if i in product_atom_maps:
+                # atom maps start at an index of 1
+                corresponding_product_atom_map = product_atom_maps.index(i) + 1
+                precursors_atom_maps.append(corresponding_product_atom_map)
+                atom.SetProp("molAtomMapNumber", str(corresponding_product_atom_map))
 
-    for i, atom in enumerate(precursors_mol.GetAtoms()):
-        if i in product_atom_maps:
-            # atom maps start at an index of 1
-            corresponding_product_atom_map = product_atom_maps.index(i) + 1
-            precursors_atom_maps.append(corresponding_product_atom_map)
-            atom.SetProp(
-                "molAtomMapNumber", str(corresponding_product_atom_map)
-            )
+                indices = [idx for idx, x in enumerate(product_atom_maps) if x == i]
 
-            indices = [idx for idx, x in enumerate(product_atom_maps) if x == i]
+                if len(indices) > 1:
+                    for idx in indices[1:]:
+                        product_mapping_dict[idx] = corresponding_product_atom_map
 
-            if len(indices) > 1:
-                for idx in indices[1:]:
-                    product_mapping_dict[idx] = corresponding_product_atom_map
+                if expected_atom_maps is not None:
+                    if (
+                        i not in expected_atom_maps
+                        or corresponding_product_atom_map
+                        != expected_atom_maps.index(i) + 1
+                    ):
+                        differing_maps.append(corresponding_product_atom_map)
+        atom_mapped_precursors_list.append(
+            Chem.MolToSmiles(precursor_mol, canonical=canonical)
+        )
 
-            if expected_atom_maps is not None:
-                if (
-                    i not in expected_atom_maps
-                    or corresponding_product_atom_map !=
-                    expected_atom_maps.index(i) + 1
-                ):
-                    differing_maps.append(corresponding_product_atom_map)
+    i = -1
+    atom_mapped_products_list = []
+    for products_mol in products_mols:
+        for atom in products_mol.GetAtoms():
+            i += 1
+            atom_map = product_mapping_dict.get(i, i + 1)
+            atom.SetProp("molAtomMapNumber", str(atom_map))
+        atom_mapped_products_list.append(
+            Chem.MolToSmiles(products_mol, canonical=canonical)
+        )
 
-    atom_mapped_precursors = Chem.MolToSmiles(precursors_mol)
-
-    for i, atom in enumerate(products_mol.GetAtoms()):
-        atom_map = product_mapping_dict.get(i, i + 1)
-        atom.SetProp("molAtomMapNumber", str(atom_map))
-    atom_mapped_products = Chem.MolToSmiles(products_mol)
-
-    atom_mapped_rxn = atom_mapped_precursors + ">>" + atom_mapped_products
+    atom_mapped_rxn = ReactionEquation(
+        atom_mapped_precursors_list, [], atom_mapped_products_list
+    )
 
     if expected_atom_maps is not None:
         return atom_mapped_rxn, differing_maps
@@ -401,88 +398,50 @@ def generate_atom_mapped_reaction_atoms(
     return atom_mapped_rxn
 
 
-class NotCanonicalizableSmilesException(ValueError):
-    pass
-
-
-def canonicalize_smi(smi: str, remove_atom_mapping=False) -> str:
-    """ Convert a SMILES string into its canonicalized form
+def canonicalize_smi(smi: str, remove_mapping: bool = False) -> str:
+    """Convert a SMILES string into its canonicalized form
 
     Args:
         smi: Reaction SMILES
-        remove_atom_mapping: If True, remove atom mapping information from the canonicalized SMILES output
+        remove_mapping: If True, remove atom mapping information from the canonicalized SMILES output
+
+    Raises:
+        InvalidSmiles: if the SMILES string cannot be canonicalized.
 
     Returns:
         SMILES reaction, canonicalized, as a string
     """
-    mol = Chem.MolFromSmiles(smi)
-    if not mol:
-        raise NotCanonicalizableSmilesException("Molecule not canonicalizable")
-    if remove_atom_mapping:
-        for atom in mol.GetAtoms():
-            if atom.HasProp("molAtomMapNumber"):
-                atom.ClearProp("molAtomMapNumber")
-    return Chem.MolToSmiles(mol)
+    if remove_mapping:
+        smi = remove_atom_mapping(smi)
+
+    return canonicalize_smiles(smi)
 
 
-def process_reaction(
-    rxn: str, fragments: str = "", fragment_bond: str = "~"
-) -> str:
+def process_reaction(reaction: ReactionEquation) -> ReactionEquation:
     """
     Remove atom-mapping, move reagents to reactants and canonicalize reaction.
-    If fragment group information is given, keep the groups together using
-    the character defined with fragment_bond.
 
     Args:
-        rxn: Reaction SMILES
-        fragments: (optional) fragments information
-        fragment_bond:
+        reaction: Reaction equation to process.
 
-    Returns: joined_precursors>>joined_products reaction SMILES 
+    Returns:
+        Processed reaction.
     """
-    reactants, reagents, products = rxn.split(">")
+    reaction = merge_reactants_and_agents(reaction)
+
     try:
-        precursors = [canonicalize_smi(r, True) for r in reactants.split(".")]
-        if len(reagents) > 0:
-            precursors += [
-                canonicalize_smi(r, True) for r in reagents.split(".")
-            ]
-        products = [canonicalize_smi(p, True) for p in products.split(".")]
-    except NotCanonicalizableSmilesException:
-        return ""
-    if len(fragments) > 1 and fragments[1] == "f":
-        number_of_precursors = len(precursors)
-        groups = fragments[3:-1].split(",")
-        new_precursors = precursors.copy()
-        new_products = products.copy()
-        for group in groups:
-            grouped_smi = []
-            if group.startswith("^"):
-                return ""
-            for member in group.split("."):
-                member = int(member)
-                if member >= number_of_precursors:
-                    grouped_smi.append(products[member - number_of_precursors])
-                    new_products.remove(
-                        products[member - number_of_precursors]
-                    )
-                else:
-                    grouped_smi.append(precursors[member])
-                    new_precursors.remove(precursors[member])
-            if member >= number_of_precursors:
-                new_products.append(fragment_bond.join(sorted(grouped_smi)))
-            else:
-                new_precursors.append(fragment_bond.join(sorted(grouped_smi)))
-        precursors = new_precursors
-        products = new_products
-    joined_precursors = ".".join(sorted(precursors))
-    joined_products = ".".join(sorted(products))
-    return f"{joined_precursors}>>{joined_products}"
+        canonicalize_and_remove_atom_map = partial(
+            canonicalize_smi, remove_mapping=True
+        )
+        reaction = apply_to_compounds(reaction, canonicalize_and_remove_atom_map)
+    except InvalidSmiles:
+        return ReactionEquation([], [], [])
+
+    reaction = sort_compounds(reaction)
+    return reaction
 
 
-def process_reaction_with_product_maps_atoms(
-    rxn,  skip_if_not_in_precursors=False
-):
+def process_reaction_with_product_maps_atoms(rxn, skip_if_not_in_precursors=False):
     """
     Remove atom-mapping, move reagents to reactants and canonicalize reaction.
     If fragment group information is given, keep the groups together using
@@ -492,25 +451,15 @@ def process_reaction_with_product_maps_atoms(
         rxn: Reaction SMILES
         skip_if_not_in_precursors: accept unmapped atoms in the product (default: False)
 
-    Returns: joined_precursors>>joined_products reaction SMILES 
+    Returns: joined_precursors>>joined_products reaction SMILES
     """
     reactants, reagents, products = rxn.split(">")
     try:
-        precursors = [
-            canonicalize_and_atom_map(r)
-            for r in reactants.split(".")
-        ]
+        precursors = [canonicalize_and_atom_map(r) for r in reactants.split(".")]
         if len(reagents) > 0:
-            precursors += [
-                canonicalize_and_atom_map(
-                    r
-                ) for r in reagents.split(".")
-            ]
-        products = [
-            canonicalize_and_atom_map(p)
-            for p in products.split(".")
-        ]
-    except NotCanonicalizableSmilesException:
+            precursors += [canonicalize_and_atom_map(r) for r in reagents.split(".")]
+        products = [canonicalize_and_atom_map(p) for p in products.split(".")]
+    except InvalidSmiles:
         return ""
     sorted_precursors = sorted(precursors, key=lambda x: x[0])
     sorted_products = sorted(products, key=lambda x: x[0])
@@ -541,9 +490,7 @@ def process_reaction_with_product_maps_atoms(
             ):  # handle equivalent atoms
                 found_alternative = False
                 for atom_idx, precursor_map in enumerate(precursors_atom_maps):
-                    if (
-                        precursor_map == p_map
-                    ) and atom_idx not in products_maps:
+                    if (precursor_map == p_map) and atom_idx not in products_maps:
                         products_maps.append(atom_idx)
                         found_alternative = True
                         break
